/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/08/22.
//

#include "grid.hpp"

#include "fggl/assets/loader.hpp"
#include "fggl/entity/gridworld/zone.hpp"

using namespace fggl::gfx::colours;

namespace demo {

	using namespace fggl::entity::grid;

	static void build_tileset(TileSet& tiles) {
		tiles.m_floors.push_back(FloorTile{FloorTile::IMPOSSIBLE, BLACK});
		tiles.m_floors.push_back(FloorTile{1, GREEN});
		tiles.m_floors.push_back(FloorTile{1, YELLOW_GREEN});

		fggl::entity::grid::WallTile noWall{};
		tiles.m_walls.push_back(noWall);

		fggl::entity::grid::WallTile solidWall{
			.render = true,
			.colour = DARK_SLATE_GRAY
		};
		tiles.m_walls.push_back(solidWall);
	}

	static void build_room(DemoGrid* area, fggl::math::vec2i center, fggl::math::vec2i size) {

		for (int yOffset = -size.y; yOffset <= size.y; ++yOffset) {
			auto yPos = yOffset + center.y;
			area->setWallAt(center.x - size.x, yPos, false, 1);
			area->setWallAt(center.x + size.x + 1, yPos, false, 1);
		}

		for (int xOffset = -size.x; xOffset <= size.x; ++xOffset) {
			auto xPos = center.x + xOffset;

			area->setWallAt(xPos, center.y - size.y, true, 1);
			area->setWallAt(xPos, center.y + size.y + 1, true, 1);

			for (int yOffset = -size.y; yOffset <= size.y; ++yOffset) {
				auto yPos = yOffset + center.y;
				area->setFloorAt(xPos, yPos, 1);
			}
		}
	}

	void build_doorway(DemoGrid* area, fggl::math::vec2i position, bool north, int size = 1) {
		for ( auto offset = 0; offset < size; offset++) {
			if ( north ) {
				area->setWallAt(position.x + offset, position.y, north, 0);
			} else {
				area->setWallAt(position.x, position.y + offset, north, 0);
			}
		}
	}

	static fggl::entity::EntityID build_test_env(DemoGrid* area) {
		area->clear();
		build_room(area, {5, 5}, {4,4});
		build_room(area, {11, 5}, {1,1});
		build_room(area, {17, 5}, {4,4});
		build_doorway(area, {10, 5}, false, 1);
		build_doorway(area, {13, 5}, false, 1);

		// set goal
		area->setFloorAt(17, 5, 2);

		build_room(area, {25, 5}, {3,3});

		// player
		fggl::entity::EntityID player = fggl::entity::INVALID;
		auto& manager = area->entities();
		{
			player = manager.create();
			auto& cellPos = manager.add<CellPos>(player);
			cellPos.pos = {5,5};
			cellPos.direction = 1;
		}
		return player;
	}

	struct Action {
		const char* name;
		std::function<void(void)> callback;
	};

	GridScene::GridScene(fggl::App &app) : GameBase(app), m_tiles(), m_animator(15.0F), m_grid(nullptr) {
		m_animator.add([this](){this->tickPlayer();});

		std::array<Action, 4> actions{{
										  {"<", [=]() { this->rotate(true); }},
										  {">", [=]() { this->rotate(false); }},
										  {"^", [=]() { this->forward(); }},
										  {"Z", [=]() {  } }
									  }};

		fggl::math::vec2i pos{0, 0};
		for (auto& action : actions) {
			fggl::math::vec2i size{32, 32};
			auto btn = std::make_unique<fggl::gui::Button>(pos, size);
			btn->label(action.name);
			btn->addCallback([=](){
				this->m_program.m_instructions.push_back(action.callback);
			});
			m_canvas.add(std::move(btn));

			if ( pos.x == 0 ) {
				pos.x += 32 + 8;
			} else {
				pos.x = 0;
				pos.y += 32 + 8;
			}
		}

		{
			fggl::math::vec2i size{32, 32};
			auto btn = std::make_unique<fggl::gui::Button>(pos, size);
			btn->label("go");
			btn->addCallback([=](){
				if ( !this->m_program.playing ) {
					this->m_program.m_currInstruction = 0;
					this->m_program.playing = true;
				}
			});
			m_canvas.add(std::move(btn));
		}

	}

	void GridScene::activate() {
		GameBase::activate();
		fggl::debug::log(fggl::debug::Level::info, "GridScene::activate()");

		m_animator.reset();

		// fake loading the tileset
		if ( m_tiles.m_floors.empty() ) {
			build_tileset(m_tiles);
		}

		// create the grid world
		m_grid = std::make_unique<DemoGrid>(m_tiles);
		m_player = build_test_env(m_grid.get());
	}

	void GridScene::deactivate() {
		m_grid = nullptr;
	}


	constexpr float DRAW_SIZE = 64.0F;
	constexpr float DRAW_HALF = DRAW_SIZE / 2.0F;
	constexpr float WALL_HALF = 2.5F;

	static void render_grid(fggl::gfx::Paint& paint, fggl::entity::grid::Area2D<255, 255>& grid) {
		const fggl::math::vec2 wallOffsetNorth {0, -DRAW_HALF};
		const fggl::math::vec2 wallOffsetWest {-DRAW_HALF, 0};

		for (int i=0; i <= 31; ++i) {
			for (int j=0; j <= 31; ++j) {
				const fggl::math::vec2 drawPos{i * DRAW_SIZE, j*DRAW_SIZE};

				// draw floor
				{
					auto &floor = grid.floorAt(i, j);
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos, {DRAW_HALF, DRAW_HALF}, floor.colour);
					paint.fill(tileGfx);
				}

				auto& wallNorth = grid.wallAt(i, j, true);
				if ( wallNorth.render ) {
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos + wallOffsetNorth, {DRAW_HALF, WALL_HALF}, wallNorth.colour);
					paint.fill(tileGfx);
				}

				auto& wallWest = grid.wallAt(i, j, false);
				if ( wallWest.render ) {
					fggl::gfx::Path2D tileGfx = fggl::gfx::make_rect(drawPos + wallOffsetWest, {WALL_HALF, DRAW_HALF}, wallWest.colour);
					paint.fill(tileGfx);
				}
			}
		}
	}


	static void render_objects(fggl::gfx::Paint& paint, DemoGrid& grid) {
		auto& manager = grid.entities();

		auto entities = manager.find<CellPos>();
		for (const auto& entity : entities ) {
			//auto& sprite = manager.get<Sprite>(entity);
			auto& cellPos = manager.get<CellPos>(entity);

			// convert grid pos to world pos
			fggl::math::vec2f drawPos = cellPos.pos;
			drawPos *= DRAW_SIZE;
			drawPos += cellPos.drawOffset;

			fggl::gfx::ShapeOpts opts;
			opts.angleOffset = (float)cellPos.direction * fggl::math::HALF_PI + cellPos.rotationOffset;

			auto shape = fggl::gfx::make_shape(drawPos, DRAW_HALF, 3, NAVAJO_WHITE, opts);
			paint.fill(shape);
		}
	}

	static void update_canvas(fggl::input::Input& inputs, fggl::gui::Container& canvas) {
		fggl::math::vec2f cursorPos {
			inputs.mouse.axis(fggl::input::MouseAxis::X),
			inputs.mouse.axis(fggl::input::MouseAxis::Y)
		};

		// in canvas space
		fggl::math::vec2 projected {
			fggl::math::rescale_ndc(cursorPos.x, 0, 1920.f),
			fggl::math::rescale_ndc(cursorPos.y, 0, 1080.0f)
		};

		auto *hoverWidget = canvas.getChildAt(projected);
		/*if (hoverWidget != m_hover) {
			if (m_hover != nullptr) {
				m_hover->onExit();
			}
			m_hover = hoverWidget;
			if (m_hover != nullptr) {
				m_hover->onEnter();
			}
		}*/

		if (inputs.mouse.pressed(fggl::input::MouseButton::LEFT)) {
			auto* widget = canvas.getChildAt(projected);
			if (widget != nullptr) {
				fggl::debug::info("Button clicked");
				widget->activate();
			}
		}
	}

	void GridScene::update(float deltaTime) {
		GameBase::update(deltaTime);
		m_animator.update(deltaTime);

		update_canvas(input(), m_canvas);
	}

	void GridScene::tickPlayer() {
		if ( !m_program.playing ){
			return;
		}

		auto &manager = m_grid->entities();
		if ( m_program.m_currInstruction < m_program.m_instructions.size() ) {
			m_program.m_instructions[ m_program.m_currInstruction ]();
			m_program.m_currInstruction++;
		} else {
			m_program.playing = false;
			m_program.m_currInstruction = 0;
			m_program.m_instructions.clear();
			checkVictory();
		}
	}

	void GridScene::checkVictory() {
		if ( !m_program.playing ) {
			auto& botPos = m_grid->entities().get<CellPos>(m_player).pos;
			auto gridCell = m_grid->floorAt(botPos.x, botPos.y);
			if ( gridCell.colour == YELLOW_GREEN ) {
				// a winner is you!
				returnToMenu();
			}
		}
	}

	void GridScene::render(fggl::gfx::Graphics &gfx) {
		fggl::gfx::Paint paint;
		render_grid(paint, *m_grid);
		render_objects(paint, *m_grid);

		m_canvas.render(paint);
		gfx.draw2D(paint);
	}

} // namespace demo
